@@ -107,26 +107,16 @@ but it will be faster as there is less recursion.
 
 import base64
 import datetime
+import json
 import math
 import re
-import sys
 import uuid
 
-if sys.version_info[:2] == (2, 6):
-    # In Python 2.6, json does not include object_pairs_hook. Use simplejson
-    # instead.
-    try:
-        import simplejson as json
-    except ImportError:
-        import json
-else:
-    import json
-
 from pymongo.errors import ConfigurationError
 
 import bson
-from bson import EPOCH_AWARE, EPOCH_NAIVE, RE_TYPE, SON
-from bson.binary import (Binary, JAVA_LEGACY, CSHARP_LEGACY, OLD_UUID_SUBTYPE,
+from bson import EPOCH_AWARE, RE_TYPE, SON
+from bson.binary import (Binary, UuidRepresentation, ALL_UUID_SUBTYPES,
                          UUID_SUBTYPE)
 from bson.code import Code
 from bson.codec_options import CodecOptions
@@ -143,13 +133,6 @@ from bson.timestamp import Timestamp
 from bson.tz_util import utc
 
 
-try:
-    json.loads("{}", object_pairs_hook=dict)
-    _HAS_OBJECT_PAIRS_HOOK = True
-except TypeError:
-    _HAS_OBJECT_PAIRS_HOOK = False
-
-
 _RE_OPT_TABLE = {
     "i": re.I,
     "l": re.L,
@@ -245,10 +228,6 @@ class JSONMode:
 class JSONOptions(CodecOptions):
     """Encapsulates JSON options for :func:`dumps` and :func:`loads`.
 
-    Raises :exc:`~pymongo.errors.ConfigurationError` on Python 2.6 if
-    `simplejson >= 2.1.0 <https://pypi.python.org/pypi/simplejson>`_ is not
-    installed and document_class is not the default (:class:`dict`).
-
     :Parameters:
       - `strict_number_long`: If ``True``, :class:`~bson.int64.Int64` objects
         are encoded to MongoDB Extended JSON's *Strict mode* type
@@ -265,9 +244,9 @@ class JSONOptions(CodecOptions):
       - `document_class`: BSON documents returned by :func:`loads` will be
         decoded to an instance of this class. Must be a subclass of
         :class:`collections.MutableMapping`. Defaults to :class:`dict`.
-      - `uuid_representation`: The BSON representation to use when encoding
-        and decoding instances of :class:`uuid.UUID`. Defaults to
-        :const:`~bson.binary.PYTHON_LEGACY`.
+      - `uuid_representation`: The :class:`~bson.binary.UuidRepresentation`
+        to use when encoding and decoding instances of :class:`uuid.UUID`.
+        Defaults to :const:`~bson.binary.UuidRepresentation.PYTHON_LEGACY`.
       - `tz_aware`: If ``True``, MongoDB Extended JSON's *Strict mode* type
         `Date` will be decoded to timezone aware instances of
         :class:`datetime.datetime`. Otherwise they will be naive. Defaults
@@ -301,11 +280,6 @@ class JSONOptions(CodecOptions):
                 "JSONOptions.datetime_representation must be one of LEGACY, "
                 "NUMBERLONG, or ISO8601 from DatetimeRepresentation.")
         self = super(JSONOptions, cls).__new__(cls, *args, **kwargs)
-        if not _HAS_OBJECT_PAIRS_HOOK and self.document_class != dict:
-            raise ConfigurationError(
-                "Support for JSONOptions.document_class on Python 2.6 "
-                "requires simplejson >= 2.1.0"
-                "(https://pypi.python.org/pypi/simplejson) to be installed.")
         if json_mode not in (JSONMode.LEGACY,
                              JSONMode.RELAXED,
                              JSONMode.CANONICAL):
@@ -430,11 +404,8 @@ def loads(s, *args, **kwargs):
        Accepts optional parameter `json_options`. See :class:`JSONOptions`.
     """
     json_options = kwargs.pop("json_options", DEFAULT_JSON_OPTIONS)
-    if _HAS_OBJECT_PAIRS_HOOK:
-        kwargs["object_pairs_hook"] = lambda pairs: object_pairs_hook(
-            pairs, json_options)
-    else:
-        kwargs["object_hook"] = lambda obj: object_hook(obj, json_options)
+    kwargs["object_pairs_hook"] = lambda pairs: object_pairs_hook(
+        pairs, json_options)
     return json.loads(s, *args, **kwargs)
 
 
@@ -478,7 +449,7 @@ def object_hook(dct, json_options=DEFAULT_JSON_OPTIONS):
     if "$code" in dct:
         return _parse_canonical_code(dct)
     if "$uuid" in dct:
-        return _parse_legacy_uuid(dct)
+        return _parse_legacy_uuid(dct, json_options)
     if "$undefined" in dct:
         return None
     if "$numberLong" in dct:
@@ -513,23 +484,32 @@ def _parse_legacy_regex(doc):
     return Regex(pattern, flags)
 
 
-def _parse_legacy_uuid(doc):
+def _parse_legacy_uuid(doc, json_options):
     """Decode a JSON legacy $uuid to Python UUID."""
     if len(doc) != 1:
         raise TypeError('Bad $uuid, extra field(s): %s' % (doc,))
-    return uuid.UUID(doc["$uuid"])
+    if json_options.uuid_representation == UuidRepresentation.UNSPECIFIED:
+        return Binary.from_uuid(uuid.UUID(doc["$uuid"]))
+    else:
+        return uuid.UUID(doc["$uuid"])
 
 
 def _binary_or_uuid(data, subtype, json_options):
     # special handling for UUID
-    if subtype == OLD_UUID_SUBTYPE:
-        if json_options.uuid_representation == CSHARP_LEGACY:
-            return uuid.UUID(bytes_le=data)
-        if json_options.uuid_representation == JAVA_LEGACY:
-            data = data[7::-1] + data[:7:-1]
-        return uuid.UUID(bytes=data)
-    if subtype == UUID_SUBTYPE:
-        return uuid.UUID(bytes=data)
+    if subtype in ALL_UUID_SUBTYPES:
+        uuid_representation = json_options.uuid_representation
+        binary_value = Binary(data, subtype)
+        if uuid_representation == UuidRepresentation.UNSPECIFIED:
+            return binary_value
+        if subtype == UUID_SUBTYPE:
+            # Legacy behavior: use STANDARD with binary subtype 4.
+            uuid_representation = UuidRepresentation.STANDARD
+        elif uuid_representation == UuidRepresentation.STANDARD:
+            # subtype == OLD_UUID_SUBTYPE
+            # Legacy behavior: STANDARD is the same as PYTHON_LEGACY.
+            uuid_representation = UuidRepresentation.PYTHON_LEGACY
+        return binary_value.as_uuid(uuid_representation)
+
     if PY3 and subtype == 0:
         return data
     return Binary(data, subtype)
@@ -573,7 +553,7 @@ def _parse_canonical_datetime(doc, json_options):
         if dtm[-1] == 'Z':
             dt = dtm[:-1]
             offset = 'Z'
-        elif dtm[-3] == ':':
+        elif dtm[-6] in ('+', '-') and dtm[-3] == ':':
             # (+|-)HH:MM
             dt = dtm[:-6]
             offset = dtm[-6:]
@@ -727,7 +707,7 @@ def _parse_canonical_decimal128(doc):
 
 def _parse_canonical_minkey(doc):
     """Decode a JSON MinKey to bson.min_key.MinKey."""
-    if doc['$minKey'] is not 1:
+    if type(doc['$minKey']) is not int or doc['$minKey'] != 1:
         raise TypeError('$minKey value must be 1: %s' % (doc,))
     if len(doc) != 1:
         raise TypeError('Bad $minKey, extra field(s): %s' % (doc,))
@@ -736,7 +716,7 @@ def _parse_canonical_minkey(doc):
 
 def _parse_canonical_maxkey(doc):
     """Decode a JSON MaxKey to bson.max_key.MaxKey."""
-    if doc['$maxKey'] is not 1:
+    if type(doc['$maxKey']) is not int or doc['$maxKey'] != 1:
         raise TypeError('$maxKey value must be 1: %s', (doc,))
     if len(doc) != 1:
         raise TypeError('Bad $minKey, extra field(s): %s' % (doc,))
@@ -823,15 +803,9 @@ def default(obj, json_options=DEFAULT_JSON_OPTIONS):
         return _encode_binary(obj, 0, json_options)
     if isinstance(obj, uuid.UUID):
         if json_options.strict_uuid:
-            data = obj.bytes
-            subtype = OLD_UUID_SUBTYPE
-            if json_options.uuid_representation == CSHARP_LEGACY:
-                data = obj.bytes_le
-            elif json_options.uuid_representation == JAVA_LEGACY:
-                data = data[7::-1] + data[:7:-1]
-            elif json_options.uuid_representation == UUID_SUBTYPE:
-                subtype = UUID_SUBTYPE
-            return _encode_binary(data, subtype, json_options)
+            binval = Binary.from_uuid(
+                obj, uuid_representation=json_options.uuid_representation)
+            return _encode_binary(binval, binval.subtype, json_options)
         else:
             return {"$uuid": obj.hex}
     if isinstance(obj, Decimal128):
