@@ -21,24 +21,14 @@ import sys
 import threading
 import collections
 
-try:
-    import ssl
-    from ssl import SSLError
-    _HAVE_SNI = getattr(ssl, 'HAS_SNI', False)
-except ImportError:
-    _HAVE_SNI = False
-    class SSLError(socket.error):
-        pass
-
-try:
-    from ssl import CertificateError as _SSLCertificateError
-except ImportError:
-    class _SSLCertificateError(ValueError):
-        pass
 
+from pymongo.ssl_support import (
+    SSLError as _SSLError,
+    HAS_SNI as _HAVE_SNI,
+    IPADDR_SAFE as _IPADDR_SAFE)
 
 from bson import DEFAULT_CODEC_OPTIONS
-from bson.py3compat import imap, itervalues, _unicode, integer_types
+from bson.py3compat import imap, itervalues, _unicode
 from bson.son import SON
 from pymongo import auth, helpers, thread_util, __version__
 from pymongo.client_session import _validate_session_write_concern
@@ -52,6 +42,7 @@ from pymongo.common import (MAX_BSON_SIZE,
                             ORDERED_TYPES,
                             WAIT_QUEUE_TIMEOUT)
 from pymongo.errors import (AutoReconnect,
+                            CertificateError,
                             ConnectionFailure,
                             ConfigurationError,
                             InvalidOperation,
@@ -65,12 +56,12 @@ from pymongo.monotonic import time as _time
 from pymongo.monitoring import (ConnectionCheckOutFailedReason,
                                 ConnectionClosedReason)
 from pymongo.network import (command,
-                             receive_message,
-                             SocketChecker)
+                             receive_message)
 from pymongo.read_preferences import ReadPreference
 from pymongo.server_type import SERVER_TYPE
+from pymongo.socket_checker import SocketChecker
 # Always use our backport so we always have support for IP address matching
-from pymongo.ssl_match_hostname import match_hostname, CertificateError
+from pymongo.ssl_match_hostname import match_hostname
 
 # For SNI support. According to RFC6066, section 3, IPv4 and IPv6 literals are
 # not permitted for SNI hostname.
@@ -131,7 +122,7 @@ except ImportError:
         """Dummy function for platforms that don't provide fcntl."""
         pass
 
-_MAX_TCP_KEEPIDLE = 300
+_MAX_TCP_KEEPIDLE = 120
 _MAX_TCP_KEEPINTVL = 10
 _MAX_TCP_KEEPCNT = 9
 
@@ -141,31 +132,36 @@ if sys.platform == 'win32':
     except ImportError:
         import winreg
 
+    def _query(key, name, default):
+        try:
+            value, _ = winreg.QueryValueEx(key, name)
+            # Ensure the value is a number or raise ValueError.
+            return int(value)
+        except (OSError, ValueError):
+            # QueryValueEx raises OSError when the key does not exist (i.e.
+            # the system is using the Windows default value).
+            return default
+
     try:
         with winreg.OpenKey(
                 winreg.HKEY_LOCAL_MACHINE,
                 r"SYSTEM\CurrentControlSet\Services\Tcpip\Parameters") as key:
-            _DEFAULT_TCP_IDLE_MS, _ = winreg.QueryValueEx(key, "KeepAliveTime")
-            _DEFAULT_TCP_INTERVAL_MS, _ = winreg.QueryValueEx(
-                key, "KeepAliveInterval")
-            # Make sure these are integers.
-            if not isinstance(_DEFAULT_TCP_IDLE_MS, integer_types):
-                raise ValueError
-            if not isinstance(_DEFAULT_TCP_INTERVAL_MS, integer_types):
-                raise ValueError
-    except (OSError, ValueError):
-        # We could not check the default values so do not attempt to override.
-        def _set_keepalive_times(dummy):
-            pass
-    else:
-        def _set_keepalive_times(sock):
-            idle_ms = min(_DEFAULT_TCP_IDLE_MS, _MAX_TCP_KEEPIDLE * 1000)
-            interval_ms = min(_DEFAULT_TCP_INTERVAL_MS,
-                              _MAX_TCP_KEEPINTVL * 1000)
-            if (idle_ms < _DEFAULT_TCP_IDLE_MS or
-                    interval_ms < _DEFAULT_TCP_INTERVAL_MS):
-                sock.ioctl(socket.SIO_KEEPALIVE_VALS,
-                           (1, idle_ms, interval_ms))
+            _WINDOWS_TCP_IDLE_MS = _query(key, "KeepAliveTime", 7200000)
+            _WINDOWS_TCP_INTERVAL_MS = _query(key, "KeepAliveInterval", 1000)
+    except OSError:
+        # We could not check the default values because winreg.OpenKey failed.
+        # Assume the system is using the default values.
+        _WINDOWS_TCP_IDLE_MS = 7200000
+        _WINDOWS_TCP_INTERVAL_MS = 1000
+
+    def _set_keepalive_times(sock):
+        idle_ms = min(_WINDOWS_TCP_IDLE_MS, _MAX_TCP_KEEPIDLE * 1000)
+        interval_ms = min(_WINDOWS_TCP_INTERVAL_MS,
+                          _MAX_TCP_KEEPINTVL * 1000)
+        if (idle_ms < _WINDOWS_TCP_IDLE_MS or
+                interval_ms < _WINDOWS_TCP_INTERVAL_MS):
+            sock.ioctl(socket.SIO_KEEPALIVE_VALS,
+                       (1, idle_ms, interval_ms))
 else:
     def _set_tcp_option(sock, tcp_option, max_value):
         if hasattr(socket, tcp_option):
@@ -279,7 +275,7 @@ def _raise_connection_failure(address, error, msg_prefix=None):
         msg = msg_prefix + msg
     if isinstance(error, socket.timeout):
         raise NetworkTimeout(msg)
-    elif isinstance(error, SSLError) and 'timed out' in str(error):
+    elif isinstance(error, _SSLError) and 'timed out' in str(error):
         # CPython 2.7 and PyPy 2.x do not distinguish network
         # timeouts from other SSLErrors (https://bugs.python.org/issue10272).
         # Luckily, we can work around this limitation because the phrase
@@ -467,6 +463,39 @@ class PoolOptions(object):
         return self.__metadata.copy()
 
 
+def _negotiate_creds(all_credentials):
+    """Return one credential that needs mechanism negotiation, if any.
+    """
+    if all_credentials:
+        for creds in all_credentials.values():
+            if creds.mechanism == 'DEFAULT' and creds.username:
+                return creds
+    return None
+
+
+def _speculative_context(all_credentials):
+    """Return the _AuthContext to use for speculative auth, if any.
+    """
+    if all_credentials and len(all_credentials) == 1:
+        creds = next(itervalues(all_credentials))
+        return auth._AuthContext.from_credentials(creds)
+    return None
+
+
+class _CancellationContext(object):
+    def __init__(self):
+        self._cancelled = False
+
+    def cancel(self):
+        """Cancel this context."""
+        self._cancelled = True
+
+    @property
+    def cancelled(self):
+        """Was cancel called?"""
+        return self._cancelled
+
+
 class SocketInfo(object):
     """Store a socket with some metadata.
 
@@ -496,23 +525,60 @@ class SocketInfo(object):
         self.enabled_for_cmap = pool.enabled_for_cmap
         self.compression_settings = pool.opts.compression_settings
         self.compression_context = None
+        self.socket_checker = SocketChecker()
+        # Support for mechanism negotiation on the initial handshake.
+        # Maps credential to saslSupportedMechs.
+        self.negotiated_mechanisms = {}
+        self.auth_ctx = {}
 
-        # The pool's pool_id changes with each reset() so we can close sockets
-        # created before the last reset.
-        self.pool_id = pool.pool_id
+        # The pool's generation changes with each reset() so we can close
+        # sockets created before the last reset.
+        self.generation = pool.generation
         self.ready = False
+        self.cancel_context = None
+        if not pool.handshake:
+            # This is a Monitor connection.
+            self.cancel_context = _CancellationContext()
+        self.opts = pool.opts
+        self.more_to_come = False
 
-    def ismaster(self, metadata, cluster_time):
+    def ismaster(self, all_credentials=None):
+        return self._ismaster(None, None, None, all_credentials)
+
+    def _ismaster(self, cluster_time, topology_version,
+                  heartbeat_frequency, all_credentials):
         cmd = SON([('ismaster', 1)])
-        if not self.performed_handshake:
-            cmd['client'] = metadata
+        performing_handshake = not self.performed_handshake
+        awaitable = False
+        if performing_handshake:
+            self.performed_handshake = True
+            cmd['client'] = self.opts.metadata
             if self.compression_settings:
                 cmd['compression'] = self.compression_settings.compressors
+        elif topology_version is not None:
+            cmd['topologyVersion'] = topology_version
+            cmd['maxAwaitTimeMS'] = int(heartbeat_frequency*1000)
+            awaitable = True
+            # If connect_timeout is None there is no timeout.
+            if self.opts.connect_timeout:
+                self.sock.settimeout(
+                    self.opts.connect_timeout + heartbeat_frequency)
 
         if self.max_wire_version >= 6 and cluster_time is not None:
             cmd['$clusterTime'] = cluster_time
 
-        ismaster = IsMaster(self.command('admin', cmd, publish_events=False))
+        # XXX: Simplify in PyMongo 4.0 when all_credentials is always a single
+        # unchangeable value per MongoClient.
+        creds = _negotiate_creds(all_credentials)
+        if creds:
+            cmd['saslSupportedMechs'] = creds.source + '.' + creds.username
+        auth_ctx = _speculative_context(all_credentials)
+        if auth_ctx:
+            cmd['speculativeAuthenticate'] = auth_ctx.speculate_command()
+
+        doc = self.command('admin', cmd, publish_events=False,
+                           exhaust_allowed=awaitable)
+        ismaster = IsMaster(doc, awaitable=awaitable)
         self.is_writable = ismaster.is_writable
         self.max_wire_version = ismaster.max_wire_version
         self.max_bson_size = ismaster.max_bson_size
@@ -521,15 +587,28 @@ class SocketInfo(object):
         self.supports_sessions = (
             ismaster.logical_session_timeout_minutes is not None)
         self.is_mongos = ismaster.server_type == SERVER_TYPE.Mongos
-        if not self.performed_handshake and self.compression_settings:
+        if performing_handshake and self.compression_settings:
             ctx = self.compression_settings.get_compression_context(
                 ismaster.compressors)
             self.compression_context = ctx
 
-        self.performed_handshake = True
         self.op_msg_enabled = ismaster.max_wire_version >= 6
+        if creds:
+            self.negotiated_mechanisms[creds] = ismaster.sasl_supported_mechs
+        if auth_ctx:
+            auth_ctx.parse_response(ismaster)
+            if auth_ctx.speculate_succeeded():
+                self.auth_ctx[auth_ctx.credentials] = auth_ctx
         return ismaster
 
+    def _next_reply(self):
+        reply = self.receive_message(None)
+        self.more_to_come = reply.more_to_come
+        unpacked_docs = reply.unpack_response()
+        response_doc = unpacked_docs[0]
+        helpers._check_command_response(response_doc, self.max_wire_version)
+        return response_doc
+
     def command(self, dbname, spec, slave_ok=False,
                 read_preference=ReadPreference.PRIMARY,
                 codec_options=DEFAULT_CODEC_OPTIONS, check=True,
@@ -542,7 +621,8 @@ class SocketInfo(object):
                 client=None,
                 retryable_write=False,
                 publish_events=True,
-                user_fields=None):
+                user_fields=None,
+                exhaust_allowed=False):
         """Execute a command or raise an error.
 
         :Parameters:
@@ -600,7 +680,7 @@ class SocketInfo(object):
         if self.op_msg_enabled:
             self._raise_if_not_writable(unacknowledged)
         try:
-            return command(self.sock, dbname, spec, slave_ok,
+            return command(self, dbname, spec, slave_ok,
                            self.is_mongos, read_preference, codec_options,
                            session, client, check, allowable_errors,
                            self.address, check_keys, listeners,
@@ -610,7 +690,8 @@ class SocketInfo(object):
                            compression_ctx=self.compression_context,
                            use_op_msg=self.op_msg_enabled,
                            unacknowledged=unacknowledged,
-                           user_fields=user_fields)
+                           user_fields=user_fields,
+                           exhaust_allowed=exhaust_allowed)
         except OperationFailure:
             raise
         # Catch socket.error, KeyboardInterrupt, etc. and close ourselves.
@@ -640,8 +721,7 @@ class SocketInfo(object):
         If any exception is raised, the socket is closed.
         """
         try:
-            return receive_message(self.sock, request_id,
-                                   self.max_message_size)
+            return receive_message(self, request_id, self.max_message_size)
         except BaseException as error:
             self._raise_connection_failure(error)
 
@@ -671,7 +751,8 @@ class SocketInfo(object):
         self.send_message(msg, max_doc_size)
         if with_last_error:
             reply = self.receive_message(request_id)
-            return helpers._check_gle_response(reply.command_response())
+            return helpers._check_gle_response(reply.command_response(),
+                                               self.max_wire_version)
 
     def write_command(self, request_id, msg):
         """Send "insert" etc. command, returning response as a dict.
@@ -687,7 +768,7 @@ class SocketInfo(object):
         result = reply.command_response()
 
         # Raises NotMasterError or OperationFailure.
-        helpers._check_command_response(result)
+        helpers._check_command_response(result, self.max_wire_version)
         return result
 
     def check_auth(self, all_credentials):
@@ -709,8 +790,7 @@ class SocketInfo(object):
                 self.authset.discard(credentials)
 
             for credentials in cached - authset:
-                auth.authenticate(credentials, self)
-                self.authset.add(credentials)
+                self.authenticate(credentials)
 
         # CMAP spec says to publish the ready event only after authenticating
         # the connection.
@@ -729,6 +809,9 @@ class SocketInfo(object):
         """
         auth.authenticate(credentials, self)
         self.authset.add(credentials)
+        # negotiated_mechanisms are no longer needed.
+        self.negotiated_mechanisms.pop(credentials, None)
+        self.auth_ctx.pop(credentials, None)
 
     def validate_session(self, client, session):
         """Validate this session before use with client.
@@ -748,18 +831,30 @@ class SocketInfo(object):
 
     def close_socket(self, reason):
         """Close this connection with a reason."""
+        if self.closed:
+            return
+        self._close_socket()
+        if reason and self.enabled_for_cmap:
+            self.listeners.publish_connection_closed(
+                self.address, self.id, reason)
+
+    def _close_socket(self):
+        """Close this connection."""
         if self.closed:
             return
         self.closed = True
-        # Avoid exceptions on interpreter shutdown.
+        if self.cancel_context:
+            self.cancel_context.cancel()
+        # Note: We catch exceptions to avoid spurious errors on interpreter
+        # shutdown.
         try:
             self.sock.close()
         except Exception:
             pass
 
-        if reason and self.enabled_for_cmap:
-            self.listeners.publish_connection_closed(
-                self.address, self.id, reason)
+    def socket_closed(self):
+        """Return True if we know socket has been closed, False otherwise."""
+        return self.socket_checker.socket_closed(self.sock)
 
     def send_cluster_time(self, command, session, client):
         """Add cluster time for MongoDB >= 3.6."""
@@ -791,7 +886,8 @@ class SocketInfo(object):
         # KeyboardInterrupt from the start, rather than as an initial
         # socket.error, so we catch that, close the socket, and reraise it.
         self.close_socket(ConnectionClosedReason.ERROR)
-        if isinstance(error, socket.error):
+        # SSLError from PyOpenSSL inherits directly from Exception.
+        if isinstance(error, (IOError, OSError, _SSLError)):
             _raise_connection_failure(self.address, error)
         else:
             raise
@@ -882,9 +978,6 @@ def _create_connection(address, options):
         raise socket.error('getaddrinfo failed')
 
 
-_PY37PLUS = sys.version_info[:2] >= (3, 7)
-
-
 def _configured_socket(address, options):
     """Given (host, port) and PoolOptions, return a configured socket.
 
@@ -905,16 +998,16 @@ def _configured_socket(address, options):
             # https://bugs.python.org/issue32185
             # We have to pass hostname / ip address to wrap_socket
             # to use SSLContext.check_hostname.
-            if _HAVE_SNI and (not is_ip_address(host) or _PY37PLUS):
+            if _HAVE_SNI and (not is_ip_address(host) or _IPADDR_SAFE):
                 sock = ssl_context.wrap_socket(sock, server_hostname=host)
             else:
                 sock = ssl_context.wrap_socket(sock)
-        except _SSLCertificateError:
+        except CertificateError:
             sock.close()
             # Raise CertificateError directly like we do after match_hostname
             # below.
             raise
-        except IOError as exc:
+        except (IOError, OSError, _SSLError) as exc:
             sock.close()
             # We raise AutoReconnect for transient and permanent SSL handshake
             # failures alike. Permanent handshake failures, like protocol
@@ -967,7 +1060,7 @@ class Pool:
 
         # Keep track of resets, so we notice sockets created before the most
         # recent reset and close them.
-        self.pool_id = 0
+        self.generation = 0
         self.pid = os.getpid()
         self.address = address
         self.opts = options
@@ -987,7 +1080,6 @@ class Pool:
 
         self._socket_semaphore = thread_util.create_semaphore(
             self.opts.max_pool_size, max_waiters)
-        self.socket_checker = SocketChecker()
         if self.enabled_for_cmap:
             self.opts.event_listeners.publish_pool_created(
                 self.address, self.opts.non_default_options)
@@ -996,7 +1088,7 @@ class Pool:
         with self.lock:
             if self.closed:
                 return
-            self.pool_id += 1
+            self.generation += 1
             self.pid = os.getpid()
             sockets, self.sockets = self.sockets, collections.deque()
             self.active_sockets = 0
@@ -1033,10 +1125,10 @@ class Pool:
     def close(self):
         self._reset(close=True)
 
-    def remove_stale_sockets(self, reference_pool_id):
+    def remove_stale_sockets(self, reference_generation, all_credentials):
         """Removes stale sockets then adds new ones if pool is too small and
-        has not been reset. The `reference_pool_id` argument specifies the
-        `pool_id` at the point in time this operation was requested on the
+        has not been reset. The `reference_generation` argument specifies the
+        `generation` at the point in time this operation was requested on the
         pool.
         """
         if self.opts.max_idle_time_seconds is not None:
@@ -1057,18 +1149,18 @@ class Pool:
             if not self._socket_semaphore.acquire(False):
                 break
             try:
-                sock_info = self.connect()
+                sock_info = self.connect(all_credentials)
                 with self.lock:
                     # Close connection and return if the pool was reset during
                     # socket creation or while acquiring the pool lock.
-                    if self.pool_id != reference_pool_id:
+                    if self.generation != reference_generation:
                         sock_info.close_socket(ConnectionClosedReason.STALE)
                         break
                     self.sockets.appendleft(sock_info)
             finally:
                 self._socket_semaphore.release()
 
-    def connect(self):
+    def connect(self, all_credentials=None):
         """Connect to Mongo and return a new SocketInfo.
 
         Can raise ConnectionFailure or CertificateError.
@@ -1084,22 +1176,21 @@ class Pool:
         if self.enabled_for_cmap:
             listeners.publish_connection_created(self.address, conn_id)
 
-        sock = None
         try:
             sock = _configured_socket(self.address, self.opts)
-        except socket.error as error:
-            if sock is not None:
-                sock.close()
-
+        except Exception as error:
             if self.enabled_for_cmap:
                 listeners.publish_connection_closed(
                     self.address, conn_id, ConnectionClosedReason.ERROR)
 
-            _raise_connection_failure(self.address, error)
+            if isinstance(error, (IOError, OSError, _SSLError)):
+                _raise_connection_failure(self.address, error)
+
+            raise
 
         sock_info = SocketInfo(sock, self, self.address, conn_id)
         if self.handshake:
-            sock_info.ismaster(self.opts.metadata, None)
+            sock_info.ismaster(all_credentials)
             self.is_writable = sock_info.is_writable
 
         return sock_info
@@ -1130,29 +1221,23 @@ class Pool:
         listeners = self.opts.event_listeners
         if self.enabled_for_cmap:
             listeners.publish_connection_check_out_started(self.address)
-        # First get a socket, then attempt authentication. Simplifies
-        # semaphore management in the face of network errors during auth.
-        sock_info = self._get_socket_no_auth()
-        checked_auth = False
+
+        sock_info = self._get_socket(all_credentials)
+
+        if self.enabled_for_cmap:
+            listeners.publish_connection_checked_out(
+                self.address, sock_info.id)
         try:
-            sock_info.check_auth(all_credentials)
-            checked_auth = True
-            if self.enabled_for_cmap:
-                listeners.publish_connection_checked_out(
-                    self.address, sock_info.id)
             yield sock_info
         except:
             # Exception in caller. Decrement semaphore.
-            self.return_socket(sock_info, publish_checkin=checked_auth)
-            if self.enabled_for_cmap and not checked_auth:
-                self.opts.event_listeners.publish_connection_check_out_failed(
-                    self.address, ConnectionCheckOutFailedReason.CONN_ERROR)
+            self.return_socket(sock_info)
             raise
         else:
             if not checkout:
                 self.return_socket(sock_info)
 
-    def _get_socket_no_auth(self):
+    def _get_socket(self, all_credentials):
         """Get or create a SocketInfo. Can raise ConnectionFailure."""
         # We use the pid here to avoid issues with fork / multiprocessing.
         # See test.test_client:TestClient.test_fork for an example of
@@ -1172,26 +1257,35 @@ class Pool:
         if not self._socket_semaphore.acquire(
                 True, self.opts.wait_queue_timeout):
             self._raise_wait_queue_timeout()
-        with self.lock:
-            self.active_sockets += 1
 
         # We've now acquired the semaphore and must release it on error.
+        sock_info = None
+        incremented = False
         try:
-            sock_info = None
+            with self.lock:
+                self.active_sockets += 1
+                incremented = True
+
             while sock_info is None:
                 try:
                     with self.lock:
                         sock_info = self.sockets.popleft()
                 except IndexError:
                     # Can raise ConnectionFailure or CertificateError.
-                    sock_info = self.connect()
+                    sock_info = self.connect(all_credentials)
                 else:
                     if self._perished(sock_info):
                         sock_info = None
+            sock_info.check_auth(all_credentials)
         except Exception:
+            if sock_info:
+                # We checked out a socket but authentication failed.
+                sock_info.close_socket(ConnectionClosedReason.ERROR)
             self._socket_semaphore.release()
-            with self.lock:
-                self.active_sockets -= 1
+
+            if incremented:
+                with self.lock:
+                    self.active_sockets -= 1
 
             if self.enabled_for_cmap:
                 self.opts.event_listeners.publish_connection_check_out_failed(
@@ -1200,38 +1294,41 @@ class Pool:
 
         return sock_info
 
-    def return_socket(self, sock_info, publish_checkin=True):
+    def return_socket(self, sock_info):
         """Return the socket to the pool, or if it's closed discard it.
 
         :Parameters:
           - `sock_info`: The socket to check into the pool.
-          - `publish_checkin`: If False, a ConnectionCheckedInEvent will not
-            be published.
         """
         listeners = self.opts.event_listeners
-        if self.enabled_for_cmap and publish_checkin:
+        if self.enabled_for_cmap:
             listeners.publish_connection_checked_in(self.address, sock_info.id)
         if self.pid != os.getpid():
             self.reset()
         else:
             if self.closed:
                 sock_info.close_socket(ConnectionClosedReason.POOL_CLOSED)
-            elif sock_info.pool_id != self.pool_id:
-                sock_info.close_socket(ConnectionClosedReason.STALE)
             elif not sock_info.closed:
-                sock_info.update_last_checkin_time()
-                sock_info.update_is_writable(self.is_writable)
                 with self.lock:
-                    self.sockets.appendleft(sock_info)
+                    # Hold the lock to ensure this section does not race with
+                    # Pool.reset().
+                    if sock_info.generation != self.generation:
+                        sock_info.close_socket(ConnectionClosedReason.STALE)
+                    else:
+                        sock_info.update_last_checkin_time()
+                        sock_info.update_is_writable(self.is_writable)
+                        self.sockets.appendleft(sock_info)
 
         self._socket_semaphore.release()
         with self.lock:
             self.active_sockets -= 1
 
     def _perished(self, sock_info):
-        """This side-effecty function checks if this socket has been idle for
+        """Return True and close the connection if it is "perished".
+
+        This side-effecty function checks if this socket has been idle for
         for longer than the max idle time, or if the socket has been closed by
-        some external network error.
+        some external network error, or if the socket's generation is outdated.
 
         Checking sockets lets us avoid seeing *some*
         :class:`~pymongo.errors.AutoReconnect` exceptions on server
@@ -1250,10 +1347,14 @@ class Pool:
         if (self._check_interval_seconds is not None and (
                 0 == self._check_interval_seconds or
                 idle_time_seconds > self._check_interval_seconds)):
-            if self.socket_checker.socket_closed(sock_info.sock):
+            if sock_info.socket_closed():
                 sock_info.close_socket(ConnectionClosedReason.ERROR)
                 return True
 
+        if sock_info.generation != self.generation:
+            sock_info.close_socket(ConnectionClosedReason.STALE)
+            return True
+
         return False
 
     def _raise_wait_queue_timeout(self):
